#include <string>
#include <iostream>
#include <cstdlib>
#include <ctime>
#include <cstring>
#include <algorithm>
#include <fstream>
#include <sstream>


#include <CL/opencl.h>

#ifdef linux
    #include <sys/time.h>
#endif

#include "util.h"




using namespace std;

void error(string str, int code)
{
    cerr << str << endl;
    exit(code);
}

void randomizeVector(std::vector<float>& vec, float min, float max)
{
    float range = max-min;
    for(size_t i=0;i<vec.size();i++)
        vec[i] = rand() * range / RAND_MAX + min;
}

int classify(float d)
{
    return d > 0.5;
}

Timer::Timer():
    running_(false), disabled_(false)
{
}


// Timer
void Timer::start()
{

    if(disabled_) return;
    #ifdef linux
        timeval tv;
        gettimeofday(&tv,NULL);
        started_ = tv.tv_sec*1.0E6 + tv.tv_usec;
    #else
        started_ = clock();
    #endif
    running_ = true;
}

double Timer::stop()
{
    if(disabled_) return 0;
    double elapsed_ = elapsed();
    running_ = false;
    return elapsed_;
}

double Timer::elapsed()
{
    if(disabled_) return 0;
    if(running_)
    {
        #ifdef linux
            timeval tv;
            gettimeofday(&tv,NULL);
            stopped_ = tv.tv_sec*1.0E6 + tv.tv_usec;
        #else
            stopped_ = clock();
        #endif
    }
    #ifdef linux
        double elapsed_ = (stopped_ - started_) / 1.0E6;
    #else
        double elapsed_ = double(stopped_ - started_) / CLOCKS_PER_SEC;
    #endif

    return elapsed_;
}

void Timer::print(string text)
{
    if(disabled_) return;
    double time = elapsed();
    cout << text << ": " << time << " s" << endl;
}

void Timer::disable()
{
    disabled_ = true;
}


/// Logger
Logger* Logger::logger_ = NULL;

Logger::Logger()
{

}

Logger* Logger::getInstance(){
  if(logger_ == NULL){
    logger_ = new Logger();
  }
  return logger_;
}


/** Returns true if line contains paramName, and value is stored in result */
bool ParseLine(string line, string paramName, double &result)
{
    paramName += ":";
    size_t found = line.find(paramName);
    if(found != string::npos)
    {
        istringstream iss(line.substr(found + paramName.length()));
        iss >> result;
        return true;
    }
    return false;
}



// openCL
void openclPrintDeviceInfo(cl_device_id device)
{
    char param_str[10240];
    cl_uint param_uint;
    //cl_platform_id param_platform_id;
    cl_device_type param_device_type;
    cl_ulong param_ulong;
    size_t param_size;
    size_t param_size_array[100];
    cl_device_local_mem_type param_local_mem_type;

    // Device name string.
    clGetDeviceInfo(device,CL_DEVICE_NAME,sizeof(param_str),param_str,NULL);
    cout << "Device name: " << param_str << endl;

    // Vendor name string.
    clGetDeviceInfo(device,CL_DEVICE_VENDOR,sizeof(param_str),param_str,NULL);
    // A unique device vendor identifier. An example of a unique device identifier could be the PCIe ID.
    clGetDeviceInfo(device,CL_DEVICE_VENDOR_ID,sizeof(param_uint),&param_uint,NULL);
    cout << "Vendor name: " << param_str << " (Vendor ID: " << param_uint << " )" << endl;

    //CL_DEVICE_PLATFORM 	Return type: cl_platform_id   The platform associated with this device.

    // The OpenCL device type. Currently supported values are one of or a combination of: CL_DEVICE_TYPE_CPU, CL_DEVICE_TYPE_GPU, CL_DEVICE_TYPE_ACCELERATOR, or CL_DEVICE_TYPE_DEFAULT.
    clGetDeviceInfo(device,CL_DEVICE_TYPE,sizeof(param_device_type),&param_device_type,NULL);
    cout << "Device type: ";
    if(param_device_type & CL_DEVICE_TYPE_CPU) cout << "CL_DEVICE_TYPE_CPU ";
    if(param_device_type & CL_DEVICE_TYPE_GPU) cout << "CL_DEVICE_TYPE_GPU ";
    if(param_device_type & CL_DEVICE_TYPE_ACCELERATOR) cout << "CL_DEVICE_TYPE_ACCELERATOR ";
    if(param_device_type & CL_DEVICE_TYPE_DEFAULT) cout << "CL_DEVICE_TYPE_DEFAULT ";
    cout << endl;

    // OpenCL version string. Returns the OpenCL version supported by the device. This version string has the following format: OpenCL<space><major_version.minor_version><space><vendor-specific information> The major_version.minor_version value returned will be 1.0.
    clGetDeviceInfo(device,CL_DEVICE_VERSION,sizeof(param_str),param_str,NULL);
    cout << "Device version: " << param_str << endl;

    // OpenCL software driver version string in the form major_number.minor_number.
    clGetDeviceInfo(device,CL_DRIVER_VERSION,sizeof(param_str),param_str,NULL);
    cout << "Driver version: " << param_str << endl;

    // Returns a space separated list of extension names
    clGetDeviceInfo(device,CL_DEVICE_EXTENSIONS,sizeof(param_str),param_str,NULL);
    replace(param_str, param_str+strlen(param_str), ' ', '\n');
    cout << "Device extensions:\n" << param_str << endl;

    // Maximum configured clock frequency of the device in MHz.
    clGetDeviceInfo(device,CL_DEVICE_MAX_CLOCK_FREQUENCY,sizeof(param_uint),&param_uint,NULL);
    cout << "Max clock frequency: " << param_uint << " MHz" << endl;

    // Size of global device memory in bytes.
    clGetDeviceInfo(device,CL_DEVICE_GLOBAL_MEM_SIZE,sizeof(param_ulong),&param_ulong,NULL);
    cout << "Global mem size: " << param_ulong/(1024*1024) << "MB" << endl;

    // Size of global memory cache in bytes.
    clGetDeviceInfo(device,CL_DEVICE_GLOBAL_MEM_CACHE_SIZE,sizeof(param_ulong),&param_ulong,NULL);
    cout << "Global mem cache size: " << param_ulong/(1024) << "kB" << endl;

    // Size of local memory arena in bytes. The minimum value is 16 KB.
    clGetDeviceInfo(device,CL_DEVICE_LOCAL_MEM_SIZE,sizeof(param_ulong),&param_ulong,NULL);
    cout << "Local mem size: " << param_ulong/(1024) << "kB" << endl;

    // Type of local memory supported. This can be set to CL_LOCAL implying dedicated local memory storage such as SRAM, or CL_GLOBAL.
    clGetDeviceInfo(device,CL_DEVICE_LOCAL_MEM_TYPE ,sizeof(param_local_mem_type),&param_local_mem_type,NULL);
    cout << "Local mem type: ";
    switch(param_local_mem_type)
    {
    case CL_LOCAL: cout << "local" << endl; break;
    case CL_GLOBAL: cout << "global" << endl; break;
    }

    // Size of local memory arena in bytes. The minimum value is 16 KB.
    clGetDeviceInfo(device,CL_DEVICE_MAX_CONSTANT_BUFFER_SIZE,sizeof(param_ulong),&param_ulong,NULL);
    cout << "Constant buffer size: " << param_ulong/(1024) << "kB" << endl;

    // The number of parallel compute cores on the OpenCL device. The minimum value is 1.
    clGetDeviceInfo(device,CL_DEVICE_MAX_COMPUTE_UNITS,sizeof(param_uint),&param_uint,NULL);
    cout << "Max compute units: " << param_uint << endl;

    // Maximum number of work-items in a work-group executing a kernel using the data parallel execution model. (Refer to clEnqueueNDRangeKernel). The minimum value is 1.
    clGetDeviceInfo(device,CL_DEVICE_MAX_WORK_GROUP_SIZE,sizeof(param_size),&param_size,NULL);
    cout << "Max work-group size: " << param_size << endl;

    // Maximum dimensions that specify the global and local work-item IDs used by the data parallel execution model. (Refer to clEnqueueNDRangeKernel). The minimum value is 3.
    clGetDeviceInfo(device,CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS,sizeof(param_uint),&param_uint,NULL);
    // Maximum number of work-items that can be specified in each dimension of the work-group to clEnqueueNDRangeKernel. Returns n size_t entries, where n is the value returned by the query for CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS. The minimum value is (1, 1, 1).
    clGetDeviceInfo(device,CL_DEVICE_MAX_WORK_ITEM_SIZES,sizeof(param_size_array),param_size_array,NULL);
    cout << "Max work-item in work-group (for each dim): (";
    for(size_t i=0;i<param_uint;i++) cout << param_size_array[i] << ", ";
    cout << ")" << endl;

}



void openclPrintAllDevicesInfo()
{
    // get platforms
    cl_uint num_entries = 100; // just get the first 100 platforms, should be enough (or call clGetPlatformIDs twice)
    cl_platform_id platforms[num_entries];
    cl_uint num_platforms;
    if(clGetPlatformIDs(num_entries,platforms,&num_platforms) != CL_SUCCESS)
        error("Error getting platform IDs", 539);

    cout << "Number of platforms: " << num_platforms << endl;

    // print platforms info
    for(size_t i=0; i<num_platforms;i++)
    {
        char param_value[10240];
        cout << "Platform " << i << endl << "=================" << endl;
        clGetPlatformInfo(platforms[i], CL_PLATFORM_NAME, 10240, param_value, NULL);
        cout << "Platform name: " << param_value << endl;
        clGetPlatformInfo(platforms[i], CL_PLATFORM_VENDOR , 10240, param_value, NULL);
        cout << "Platform vendor: " << param_value << endl;
        clGetPlatformInfo(platforms[i], CL_PLATFORM_VERSION, 10240, param_value, NULL);
        cout << "Platform version: " << param_value << endl;
        clGetPlatformInfo(platforms[i], CL_PLATFORM_EXTENSIONS, 10240, param_value, NULL);
        cout << "Extensions: " << param_value << endl;

        cl_device_id devices[1000];
        cl_uint num_devices;
        clGetDeviceIDs(platforms[i],CL_DEVICE_TYPE_ALL,10,devices,&num_devices);
        cout << "\nNumber of devices:" << num_devices << endl;
        for(size_t j=0;j<num_devices;j++)
        {
            cout << "Device " << j << endl << "-----------------" << endl;
            openclPrintDeviceInfo(devices[j]);
        }
    }

}

/** Returns a context with the first device of type device_type it finds.
    Also returns the device. */
cl_context openclCreateContextFromType(cl_device_type device_type,cl_device_id *device)
{
    // get platforms
    cl_uint num_entries = 100; // just get the first 100 platforms, should be enough (or call clGetPlatformIDs twice)
    cl_platform_id platforms[num_entries];
    cl_uint num_platforms;
    if(clGetPlatformIDs(num_entries,platforms,&num_platforms) != CL_SUCCESS)
        error("Error getting platform IDs", 538);


    for(size_t i=0; i<num_platforms;i++)
    {
        cl_int err;
        cl_device_id devices[1000];
        cl_uint num_devices;
        clGetDeviceIDs(platforms[i],device_type,1000,devices,&num_devices);
        if(num_devices > 0)
        {
            cl_context context;
            cl_context_properties props[3];
            props[0] = (cl_context_properties)CL_CONTEXT_PLATFORM;
            props[1] = (cl_context_properties)platforms[i];
            props[2] = (cl_context_properties)0;
            context = clCreateContextFromType(props, device_type, NULL, NULL, &err);
            if(err != CL_SUCCESS)
                error("Error creating context", 550);
            *device = devices[0];

            return context;
        }
    }

    error("No devices of the type found", 973);
    return NULL;
}

string readFile(string fileName)
{
    ifstream file(fileName.c_str());
    if(!file.is_open())
        error("Error opening program file.",356);

    stringstream ss;
    ss << file.rdbuf();
    file.close();
    return ss.str();
}

cl_program openclCreateProgramFromFile(cl_context& context, string fileName)
{
    string source = readFile(fileName);

    cl_program program = openclCreateProgramFromSource(context,source);

    return program;
}

cl_program openclCreateProgramFromSource(cl_context& context, string source)
{
    const char *c_source = source.c_str();
    size_t length = source.length();

    cl_int err;
    cl_program program = clCreateProgramWithSource(context,1,&c_source,&length, &err);
    if(err != CL_SUCCESS)
        error("Error creating cl_program.",721);

    return program;
}

void openclBuildProgram(cl_program& program, cl_device_id& device)
{
    cl_int err = clBuildProgram (program, 0, NULL, 0, NULL, NULL); // "-cl-fast-relaxed-math"
    if(err != CL_SUCCESS)
    {
        char* log;
        size_t log_size;
        clGetProgramBuildInfo(program,device,CL_PROGRAM_BUILD_LOG,0,NULL,&log_size);
        log = new char[log_size+1];
        clGetProgramBuildInfo(program,device,CL_PROGRAM_BUILD_LOG,log_size,log,NULL);
        log[log_size] = '\0';
        cout << log << endl;
        delete[] log;
        error("Build error",527);
    }
}

string findAndReplace(string input, string find, string replace)
{
    string out;

    size_t pos;

    while( (pos = input.find(find)) != string::npos)
    {
        out.append(input.substr(0,pos));
        input.erase(0,pos+find.length());
        out.append(replace);
    }
    out.append(input);

    return out;
}
